module Rookout
  module ComWs
    require "openssl"
    require "websocket/driver"
    require "uri"

    require_relative "../config"
    require_relative "../logger"
    require_relative "../exceptions"

    class WebsocketClient
      def initialize url, proxy, token
        @token = token
        @connection = WebsocketConnection.new url, proxy
        @proxy = proxy
        @driver = nil
      end

      def connect
        connection_error = nil
        @driver = WebSocket::Driver.client @connection

        headers.each do |key, value|
          @driver.set_header key, value
        end

        @driver.on :error do |error|
          connection_error = error
        end

        # Connect to the remote server
        # TODO: ADD CONNECT TIMEOUT
        @connection.connect @driver
        @driver.start

        while @driver.state == :connecting
          recv_data = @connection.read_char
          @driver.parse recv_data
        end

        raise Exceptions::RookWebsocketException, connection_error if @driver.state != :open
      end

      def connection_pump message_handler
        @driver.on :message do |e|
          message_handler.call e.data
        end

        until @driver.state == :closed
          recv_data = @connection.read_char
          @driver.parse recv_data
        end
      end

      def send_frame msg
        @driver.binary msg
      end

      def ping message, &callback
        @driver.ping message, &callback
      end

      def close
        return if @driver.nil?

        begin
          @driver.close
        rescue RuntimeError, Errno::EPIPE
          # Protocol close may fail if the connection is already closed
          nil
        end
        @connection.close
      end

      private

      def headers
        result = {
          "User-Agent" => "RookoutAgent/#{Config.rookout_version}+#{Config.rookout_commit}"
        }

        result["X-Rookout-Token"] = @token unless @token.nil?

        result
      end
    end

    class WebsocketConnection
      def initialize url, proxy
        @url = url
        @proxy = proxy

        @uri = URI.parse @url
        @secure = %w[https wss].include? @uri.scheme

        @socket = nil

        @proxy_connected = false
      end

      def connect driver
        @socket = tcp_connect

        proxy_connect driver if @proxy

        @socket = ssl_connect @socket if @secure
      end

      def proxy_connect driver
        connection_error = nil
        @proxy_connected = false
        proxy = driver.proxy @proxy

        proxy.on :connect do
          @proxy_connected = true
        end

        proxy.on :error do |error|
          connection_error = error
        end

        proxy.start

        proxy.parse read_char until @proxy_connected == true || !connection_error.nil?

        raise Exceptions::RookProxyException, connection_error unless connection_error.nil?
      end

      attr_reader :url

      def read_char
        @socket.getc
      end

      def write buffer
        @socket << buffer
      end

      def close
        return if @socket.nil?

        @socket.close
        @socket = nil
      end

      private

      def tcp_connect
        if @proxy
          uri = URI(@proxy)
          TCPSocket.new uri.host, uri.port
        else
          TCPSocket.new @uri.host,
                        @uri.port || (@secure ? 443 : 80)
        end
      end

      def ssl_connect tcp_socket
        ctx = ::OpenSSL::SSL::SSLContext.new
        ctx.min_version = ::OpenSSL::SSL::TLS1_2_VERSION
        cert_store = ::OpenSSL::X509::Store.new
        cert_store.set_default_paths
        ctx.cert_store = cert_store

        ssl_socket = ::OpenSSL::SSL::SSLSocket.new tcp_socket, ctx
        ssl_socket.hostname = @uri.host
        ssl_socket.sync_close = true
        ssl_socket.connect

        ssl_socket
      end
    end
  end
end
