module Rookout
  module Exceptions
    class ToolException < RuntimeError
      def initialize msg, parameters = {}, cause = nil
        super msg
        @parameters = parameters
        @cause = cause
      end

      attr_reader :parameters
    end

    class RookInterfaceException < ToolException
      def initialize message
        super message
      end
    end

    class RookVersionNotSupported < ToolException
      def initialize parameter, value
        super "Rook is not supported in this #{parameter}: #{value}"
      end
    end

    class RookAugInvalidKey < ToolException
      def initialize key, configuration
        super "Failed to get key #{key} from configuration #{configuration}",
              { key: key, configuration: configuration }
      end
    end

    class RookAttributeNotFound < ToolException
      def initialize attribute
        super "Failed to get attribute #{attribute}", { "attribute" => attribute }
      end
    end

    class RookKeyNotFound < ToolException
      def initialize key
        super "Failed to get key #{key}", { key: key }
      end
    end

    class RookWriteAttributeNotSupported < ToolException
      def initialize namespace_type, attribute
        super "Namespace #{namespace_type} does not support write",
              { namespace: namespace_type, attribute: attribute }
      end
    end

    class RookMethodNotFound < ToolException
      def initialize namespace_type, method
        super "Namespace method not found #{method}",
              { namespace: namespace_type, method: method }
      end
    end

    class RookInvalidObjectForAccess < ToolException
      def initialize object_type, access
        super "Object #{object_type} does not support access #{access}",
              { object: object_type, access: access }
      end
    end

    class RookInvalidArithmeticPath < ToolException
      def initialize path, cause = nil
        super "Invalid arithmetic path configuration. configuration: #{path}, innerException: #{cause}",
              { configuration: path },
              cause
      end
    end

    class RookOperationReadOnly < ToolException
      def initialize type
        super "Operation does not support write: #{type}", { operation: type }
      end
    end

    class RookExceptionEvaluationFailed < ToolException
      def initialize expression, cause
        super "Failed to evaluate expression %s: #{expression}", { expression: expression }, cause
      end
    end

    class RookNonPrimitiveObjectType < ToolException
      def initialize path
        super "Object #{path} must be of primitive type, such as: string, int, long etc",
              { "path" => path }
      end
    end

    class RookMessageSizeExceeded < ToolException
      def initialize message_size, max_message_size
        super "Message size of #{message_size} exceeds max size limit of #{max_message_size}. " \
                  "Change the depth of collection or change the default by setting ROOKOUT_MAX_MESSAGE_SIZE " \
                  "as environment variable or system property",
              { "message_size" => message_size, "max_message_size" => max_message_size }
      end
    end

    class RookRuleRateLimited < ToolException
      def initialize
        super "Breakpoint was disabled due to rate-limiting. " \
                   "For more information: https://docs.rookout.com/docs/breakpoints-tasks.html#rate-limiting"
      end
    end

    class RookInvalidToken < ToolException
      def initialize token = ""
        super "The Rookout token supplied #{token[0..6]} is not valid; please check the token and try again",
              { token: token[0...6] }
      end
    end

    class RookSourceFilePathSuggestion < ToolException
      def initialize wanted_path, matching_path
        super "Rookout found alternative file path: #{matching_path}",
              { "wanted_path" => wanted_path, "matching_path" => matching_path }
      end
    end

    class RookSetTracepointFailed < ToolException
      def initialize target_line, error
        super "Failed to set TracePoint on line #{target_line} due to #{error.message}",
              { target_line: target_line, error: error }
      end
    end

    class RookObjectCannotBeSerialized < ToolException
      def initialize object, message
        super message, { "class" => object.class.to_s }
      end
    end

    class RookMissingToken < ToolException
      def initialize
        super "No Rookout token was supplied. Make sure to pass the Rookout Token when starting the rook"
      end
    end

    class RookInvalidOptions < ToolException
      def initialize description
        super description
      end
    end

    class RookInvalidLabel < ToolException
      def initialize label_name
        super "Invalid label: must not start with the '$' character (#{label_name})"
      end
    end

    class RookCrcMismatchException < ToolException
      def initialize filepath, expected, calculated
        super "Line CRC32s do not match! path: #{filepath}, expected: #{expected}, calculated:#{calculated}",
              {
                "filepath" => filepath,
                "expected" => expected,
                "calculated" => calculated
              }
      end
    end

    class RookLineMoved < ToolException
      def initialize filepath, old_line_no, new_line_no
        super "Line has moved! path: #{filepath}, original line no: #{old_line_no}, new line no: #{new_line_no}",
              {
                "filepath" => filepath,
                "old_line_no" => old_line_no,
                "new_line_no" => new_line_no
              }
      end
    end

    class RookCommunicationException < ToolException
      def initialize
        super "Failed to connect to the controller - will continue attempting in the background"
      end
    end

    class RookWebsocketException < ToolException
      def initialize error
        super "Error from Websocket #{error}", { "error" => error }
      end
    end

    class RookProxyException < ToolException
      def initialize error
        super "Error from proxy #{error}", { "error" => error }
      end
    end

    class RookBadProtobuf < ToolException
      def initialize
        super 'Bad protobuf version. Please execute "bundle config force_ruby_platform true" before "bundler install".'
      end
    end

    class RookBadProtobufPlatform < ToolException
      def initialize platform
        super "Bad protobuf platform: #{platform}"
      end
    end

    class RookObjectNameMissing < ToolException
      def initialize configuration
        super "Failed to find object name",
              {
                "configuration" => configuration
              }
      end
    end

    class RookUnsupportedLocation < ToolException
      def initialize location
        super "Unsupported aug location was specified: #{location}",
              {
                "location" => location
              }
      end
    end
  end
end
