module Rookout
  require "fileutils"
  require "English"

  require_relative "config"

  class Logger
    require "singleton"
    include Singleton

    LOG_LEVELS = [:DEBUG, :INFO, :WARNING, :ERROR].freeze

    def initialize
      # Detect unit tests
      if (Config.debug ||
          $PROGRAM_NAME.end_with?("minitest_runner.rb") ||
          $PROGRAM_NAME.end_with?("tunit_or_minitest_in_folder_runner.rb")) &&
         Dir.pwd.end_with?("ruby-sdk")
        Config.logger_log_level = :DEBUG
        Config.logger_log_to_stderr = true
      end

      @verbosity = LOG_LEVELS.index(Config.logger_log_level) || LOG_LEVELS.index(:INFO)

      @output = nil
      @handlers = []
      build_handlers
    end

    def register_output output
      @output = output
    end

    def remove_output _output
      @output = nil
    end

    def debug message, *args
      log :DEBUG, message, args
    end

    def info message, *args
      log :INFO, message, args
    end

    def warning message, *args
      log :WARNING, message, args
    end

    def error message, *args
      log :ERROR, message, args
    end

    def exception message, exc
      error message, exc
    end

    private

    class LogRecord
      def initialize level, message, arguments
        @level = level
        @time = Time.new
        @message = message
        @formatted_message = @message % arguments

        arguments.each do |argument|
          if argument.is_a? Exception
            @formatted_message += "\n" + argument.message + "\n" + argument.backtrace.join("\n\t")
          end
        end

        @arguments = arguments
        set_caller
      end

      attr_reader :level, :time, :filename, :lineno, :message, :formatted_message, :arguments

      def format
        "#{@time} #{Process.pid}:#{Thread.current.name}-" \
          "#{File.basename @filename}:#{@label}@#{lineno} - #{@level} - #{formatted_message}"
      end

      private

      def set_caller
        @filename = nil
        @lineno = nil
        @label = nil

        caller_locations(4, 7).each do |caller|
          next if caller.absolute_path.end_with? "lib/rookout/logger.rb"

          @filename = caller.path
          @lineno = caller.lineno
          @label = caller.label
          break
        end
      end
    end

    def log level, message, args
      level_no = LOG_LEVELS.index level
      if level_no.nil?
        level = :ERROR
        level_no = LOG_LEVELS.index level
      end

      return if level_no < @verbosity

      record = LogRecord.new level, message, args
      @handlers.each { |handler| handler.call record }
    end

    def build_handlers
      if Config.logger_log_to_stderr
        @handlers.push new_file_handler unless Config.logger_filename.nil? || Config.logger_filename.empty?
        @handlers.push new_stderr_handler
      end

      @handlers.push new_remote_handler
    end

    def new_file_handler
      log_file_path = calculate_log_file_path
      directory = File.dirname log_file_path

      begin
        FileUtils.mkdir_p directory
        file = File.open log_file_path, "a"
      rescue SystemCallError => e
        file = nil

        if Config.debug
          STDERR.puts "[Rookout] Failed to open log file: #{log_file_path}"
          STDERR.puts e.backtrace
        end
      end

      ->(record) { file.write record.format + "\n" if file }
    end

    def calculate_log_file_path
      return Config.logger_filename if absolute_path? Config.logger_filename

      if RUBY_PLATFORM.include? "darwin"
        File.join ENV["HOME"], Config.logger_filename
      elsif RUBY_PLATFORM.match?(/cygwin|mswin|mingw|bccwin|wince|emx/)
        File.join ENV["USERPROFILE"], Config.logger_filename
      else
        File.join "/var/log", Config.logger_filename
      end
    end

    def new_stderr_handler
      ->(record) { STDERR.puts record.format }
    end

    def new_remote_handler
      lambda do |record|
        return unless @output
        @output.send_log_message record.level,
                                 record.time,
                                 record.filename,
                                 record.lineno,
                                 record.message,
                                 record.formatted_message,
                                 record.arguments
      end
    end

    def absolute_path? path
      path == File.absolute_path(path)
    end
  end
end
