module Rookout
  module Processor
    module Namespaces
      require "bigdecimal"
      require "google/protobuf/well_known_types"

      require_relative "../../exceptions"

      require_relative "namespace"
      require_relative "ruby_object_serializer"

      class ObjectDumpConfig
        def initialize max_depth, max_width, max_collection_depth, max_string
          @max_depth = max_depth
          @max_width = max_width
          @max_collection_depth = max_collection_depth
          @max_string = max_string
        end

        attr_reader :max_depth, :max_width, :max_collection_depth, :max_string
      end

      OBJECT_DUMP_CONFIG_STRICT = ObjectDumpConfig.new 2, 10, 1, 128
      OBJECT_DUMP_CONFIG_DEFAULT = ObjectDumpConfig.new 4, 20, 2, 512
      OBJECT_DUMP_CONFIG_TOLERANT = ObjectDumpConfig.new 5, 50, 4, 4 * 1024

      OBJECT_DUMP_CONFIG_STRING = ObjectDumpConfig.new 1, 0, 0, 64 * 1024
      OBJECT_DUMP_CONFIG_COLLECTION = ObjectDumpConfig.new 4, 100, 2, 512

      OBJECT_DUMP_TABLE = {
        "" => OBJECT_DUMP_CONFIG_DEFAULT,
        "strict" => OBJECT_DUMP_CONFIG_STRICT,
        "default" => OBJECT_DUMP_CONFIG_DEFAULT,
        "tolerant" => OBJECT_DUMP_CONFIG_TOLERANT
      }.freeze

      class RubyObjectNamespace < Namespace
        include RubyObjectSerializer

        def initialize obj, dump_config = OBJECT_DUMP_CONFIG_DEFAULT
          @obj = obj
          @dump_config = dump_config
        end

        attr_reader :obj, :dump_config

        def tailor_limits!
          if @obj.is_a? String
            @dump_config = OBJECT_DUMP_CONFIG_STRING
          elsif @obj.is_a?(Hash) || @obj.is_a?(Array)
            @dump_config = OBJECT_DUMP_CONFIG_COLLECTION
          else
            @dump_config = OBJECT_DUMP_CONFIG_TOLERANT
          end
          self
        end

        def read_attribute name
          attributes = @obj.instance_variables
          raise RookAttributeNotFound, name unless attributes.include? name.to_sym
          RubyObjectNamespace.new @obj.instance_variable_get(name)
        end

        def read_key key
          if @obj.is_a? Array
            key_int = key.to_i
            return RubyObjectNamespace.new @obj[key_int] if key_int >= 0 && key_int < @obj.length
            raise Exceptions::RookKeyNotFound, key

          elsif @obj.is_a? Hash
            return RubyObjectNamespace.new @obj[key] if @obj.key? key
            return RubyObjectNamespace.new @obj[key.to_sym] if key.is_a?(String) && @obj.key?(key.to_sym)

            @obj.each { |it, value| return RubyObjectNamespace.new value if it.to_s == key }

            raise Exceptions::RookKeyNotFound, key
          else
            raise Exceptions::RookInvalidObjectForAccess.new(obj.class.to_s, "ReadKey")
          end
        end

        def call_method name, args
          case name
          when "type"
            RubyObjectNamespace.new @obj.class.to_s
          when "size"
            RubyObjectNamespace.new @obj.length
          else
            super
          end
        end

        def dump log_object_errors
          dump_raw_object @obj, 0, @dump_config, log_object_errors
        end
      end
    end
  end
end
