module Rookout
  module Processor
    module Paths
      module Canopy
        require_relative "consts"

        require_relative "../../../exceptions"
        require_relative "../../namespaces/container_namespace"

        class Marker; end

        class Operation < Marker
          def read _namespace, _create
            raise NotImplementedError
          end

          def write _namespace, _value
            raise Exception::RookOperationReadOnly, self.class.to_s
          end
        end

        class FunctionOperation < Operation
          def initialize name, args
            @name = name
            @args = args
          end

          def read namespace, _create
            namespace.call_method @name, @args
          end
        end

        class AttributeOperation < Operation
          def initialize name
            @name = name
          end

          def read namespace, create
            namespace.read_attribute @name
          rescue Exceptions::RookAttributeNotFound
            raise unless create
            namespace.write_attribute @name, Namespaces::ContainerNamespace.new
          end

          def write namespace, value
            namespace.write_attribute @name, value
          end
        end

        class LookupOperation < Operation
          def initialize text
            if (text.start_with?("'") && text.end_with?("'")) || (text.start_with?("\"") && text.end_with?("\""))
              @key = text[1...-1]
            else
              @key = text.to_i
            end
          end

          def read namepsace, _create
            namepsace.read_key @key
          end
        end

        class ObjectMarker < Marker
          def initialize text, obj
            @text = text
            @obj = obj
          end

          attr_reader :text, :obj

          def to_s
            @obj.to_s
          end
        end

        class ToolExceptionMarker < ObjectMarker
          def initialize exc
            super exc.to_s, exc
          end
        end

        class NamespaceResult < ObjectMarker
          def initialize namespace, text
            super text, namespace
          end
        end

        class Text < ObjectMarker
          def initialize text
            super text, text
          end
        end

        class TextResult < ObjectMarker
          def initialize text
            super text, text
          end
        end

        class List < ObjectMarker
          def initialize list, text
            super text, list
          end
        end

        class FloatNumber < ObjectMarker
          def initialize text
            super text, text.to_f
          end
        end

        class Number < ObjectMarker
          def initialize text
            super text, text.to_i
          end
        end

        class Char < ObjectMarker
          def initialize text
            super "'@{text}'", text
          end
        end

        class Bool < ObjectMarker
          def initialize text
            super text, %w[true True].include?(text)
          end
        end

        class Nil < ObjectMarker
          def initialize
            super text, obj
          end
        end

        class Opt < Marker
          def initialize opt
            @opt = opt
            @level = nil

            # Some ops have alternatives we convert back to standard
            @opt = OPS_ALTERNATIVES[@opt.upcase] if OPS_ALTERNATIVES.key? @opt.upcase

            ALL_LEVELS.each_with_index do |ops, index|
              if ops.include? @opt
                @level = index
                break
              end
            end

            raise Exceptions::RookInvalidArithmeticPath, @opt if @level.nil?
          end

          attr_reader :level

          def check_result result, first_obj, second_obj
            return unless [true, false].include? result
            if result
              return
            end

            unless first_obj.nil?
              raise Exceptions::RookNonPrimitiveObjectType, first_obj.class.to_s unless
                PRIMITIVES.include? first_obj.class
            end

            return if second_obj.nil?
            raise Exceptions::RookNonPrimitiveObjectType, second_obj.class.to_s unless
              PRIMITIVES.include? second_obj.class
          end

          def execute_operation first, second
            # Remove  wrapping of objects as needed
            first_obj = decapsulate_item first
            second_obj = decapsulate_item second

            result = nil
            begin
              result = OPS_FUNCTIONS[@opt].call first_obj, second_obj
            rescue StandardError => e
              raise Exceptions::RookExceptionEvaluationFailed.new "", e
            end

            check_result result, first_obj, second_obj

            # If we don't have a result
            if result.nil?
              # Verify objects are primitives
              raise Exceptions::RookNonPrimitiveObjectType, a.text unless PRIMITIVES.include? first_obj.class
              raise Exceptions::RookNonPrimitiveObjectType, b.text unless PRIMITIVES.include? second_obj.class
            end

            ObjectMarker.new result.to_s, result
          end

          private

          def decapsulate_item item
            return item.obj.obj if item.is_a? NamespaceResult
            item.obj
          end
        end
      end
    end
  end
end
