module Rookout
  module Services
    require "zlib"

    require_relative "../logger"
    require_relative "../exceptions"

    class PositionMarker
      def initialize lineno, method
        @lineno = lineno
        @method = method
      end

      attr_reader :lineno, :method
    end

    class PositionResolver
      def initialize tracer
        @tracer = tracer
        @augs = {}
        @iseqs = []
        @trace_point = TracePoint.new :script_compiled do |tp|
          begin
            begin
              iseq = tp.instruction_sequence
              # Ignore script without sources
              if iseq.absolute_path
                @iseqs << iseq
                evaluate_script iseq
              end
            rescue Exception => e
              Logger.instance.exception "Exception while evaluating script", e
            end
          rescue
          end
        end

        @trace_point.enable
      end

      def add_aug location
        positions = evaluate_all_scripts_to_location location
        @tracer.add_breakpoint_aug positions, location unless positions.empty?
        @augs[location.id] = location
      end

      def remove_aug aug_id
        location = @augs[aug_id]
        return if location.nil?

        @augs.delete [aug_id]
        location.notify_removed
      end

      def clear_augs
        @augs.values.each do |aug_id|
          remove_aug aug_id
        end

        @augs = {}
      end

      def close
        clear_augs
        @trace_point.disable
      end

      private

      #########################################################################################
      # ISEQ comparison Section
      def evaluate_script iseq
        filename = iseq.absolute_path
        return if filename.nil?

        @augs.each_value do |location|
          position = evaluate_script_to_location iseq, filename, location
          next if position.nil?

          @tracer.add_breakpoint_aug [position], location
        end
      end

      def evaluate_all_scripts_to_location location
        positions = []
        @iseqs.each do |iseq|
          position = evaluate_script_to_location iseq, iseq.absolute_path, location
          next if position.nil?

          positions << position
        end

        positions
      end

      def evaluate_script_to_location iseq, filename, location
        if exact_match? location.filename, filename
          lineno = find_updated_line filename, location
          return if lineno == -1

          PositionMarker.new lineno, iseq
        elsif suggested_match? location, filename
          warning = Exceptions::RookSourceFilePathSuggestion.new location.filename, filename
          location.notify_warning warning
          # Must return nil in order to skip this script
          nil
        end
      end

      #########################################################################################
      # Utils
      def exact_match? location_filename, script_filename
        return unless script_filename.end_with? location_filename
        File.basename(script_filename) == File.basename(location_filename)
      end

      def suggested_match? location, filename
        return unless File.basename(location.filename) == File.basename(filename)
        location.file_hash == get_hash(filename)
      end

      def get_hash filename
        content = File.read filename
        content.gsub!(/(?:\r\n|\r|\n)/, "\n")
        Digest::SHA2.new(256).hexdigest content
      end

      def handle_unique_line lines, location, filename
        first_line = nil
        second_found = false

        lines.each_with_index do |line, index|
          if crc_line(line) == location.line_crc
            if first_line.nil?
              first_line = index
            else
              second_found = true
              break
            end
          end
        end

        if first_line && !second_found
          updated_line = first_line + 1
          location.notify_warning Processor::RookError.new Exceptions::RookLineMoved.new(filename,
                                                                                         location.lineno,
                                                                                         updated_line)
          return updated_line
        end
        nil
      end

      def find_updated_line filename, location
        return location.lineno if location.line_crc.nil?

        lines = File.readlines filename, chomp: true
        line_crc32 = lines.length >= location.lineno ? crc_line(lines[location.lineno - 1]) : nil
        return location.lineno if location.line_crc == line_crc32

        if location.line_unique
          updated_line = handle_unique_line lines, location, filename
          unless updated_line.nil?
            return updated_line
          end
        end

        location.notify_error Exceptions::RookCrcMismatchException.new(filename, location.line_crc, line_crc32)
        -1
      end

      def crc_line line
        (Zlib.crc32(line) & 0xffffffff).to_s 16
      end
    end
  end
end
